import imaplib
from typing import Union
from .folder import Folder, FoldList
from .utils import imap_utf7_encode, imap_utf7_decode


class ImapEasyBox:
    server: Union[imaplib.IMAP4, imaplib.IMAP4_SSL, None]

    def __init__(self, host: str, port=993, user: str | None = None, password: str | None = None, ssl: bool = True,
                 **kwargs):
        self.host = host
        self.port = port
        self.user = user
        self.password = password
        self.imap_cls = getattr(imaplib, 'IMAP4_SSL') if ssl else getattr(imaplib, 'IMAP4')
        self.server = None
        self.kwargs = kwargs
        # self._folders是邮箱中文名和原始名称构成的字典
        self._folders = None

    def login(self, user: str | None = None, password: str | None = None):
        """登陆并获取所有文件夹名称"""

        if user is None:
            user = self.user
        if password is None:
            password = self.password

        self.server = self.imap_cls(self.host, self.port, **self.kwargs)
        # 登录成果返回('OK', [b'LOGIN completed'])
        # 用户名密码错误抛出异常imaplib.IMAP4.error: b'LOGIN failure, invalid username/password'
        # 邮箱地址错误抛出异常imaplib.IMAP4.error: LOGIN command error: BAD [b'LOGIN failure, domain is disable.']
        self.server.login(user, password)
        self._update_folders()

    def quit(self):
        # 需要先选择select邮箱，然后再close，否则会抛出错误
        try:
            self.server.close()
        except imaplib.IMAP4.error:
            pass

        self.server.logout()

    def __getattr__(self, item: str):
        return getattr(self.server, item)

    def __enter__(self):
        self.login(self.user, self.password)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.quit()

    def select(self, folder_name: str) -> Folder:
        folder_raw_name = self._folders[folder_name.lower()]
        self.server.select(folder_raw_name)
        return Folder(folder_name, self)

    @property
    def folders(self) -> FoldList:
        return FoldList(Folder(folder_name, self) for folder_name in self._folders.keys())

    def _update_folders(self):
        """更新文件夹列表"""
        self._folders = {}

        # list返回的结果是('OK', [b'(\\Marked) "/" "INBOX"', b'(\\Marked) "/" "&XfJT0ZAB-"'])
        typ, data = self.server.list()

        # 结果中类似&XfJT0ZAB-的字符串是utf7编码，并且把+号替换回&符号
        for folder_byte in data:
            folder_name_byte = folder_byte.split(b' ')[-1].replace(b'"', b'')
            folder_name_key = imap_utf7_decode(folder_name_byte).lower()
            folder_name_val = folder_name_byte.decode('ascii')
            self._folders[folder_name_key] = folder_name_val

    def create_folder(self, folder_name: str):
        """创建文件夹"""
        # 创建已存在的文件夹返回('NO', [b'CREATE Folder exist']
        folder_name = imap_utf7_encode(folder_name)
        self.server.create(folder_name)
        self._update_folders()

    def rename_folder(self, old_folder_name: str, new_folder_name: str):
        """修改文件夹名称"""
        try:
            old_folder_name = self._folders[old_folder_name.lower()]
        except KeyError:
            raise NameError(f"Folder<{old_folder_name}>不存在")

        new_folder_name = imap_utf7_encode(new_folder_name).decode('ascii')
        self.server.rename(old_folder_name, new_folder_name)
        self._update_folders()

    def delete_folder(self, folder_name: str):
        """删除文件夹"""
        # 删除不存在的文件夹会返回('NO', [b'DELETE Folder not exist'])
        try:
            folder_name = self._folders[folder_name.lower()]
        except KeyError:
            raise NameError(f"Folder<{folder_name}>不存在")

        self.server.delete(folder_name)
        self._update_folders()



from email.message import Message
from email.header import decode_header
from pathlib import Path
from typing import Union
import base64


def decode_mail_header(header):
    """解析邮件元数据"""

    results = decode_header(header)
    values = []

    for val, encoding in results:
        if encoding is None:
            encoding = 'utf8'

        if isinstance(val, bytes):
            val = val.decode(encoding)

        values.append(val)

    return values


def parse_raw_mail(raw_mail: Message):
    """递归解析原始邮件"""

    content = {
        "text_body": None,
        "html_body": None,
        "html_encoding": None,
        "attachments": [],
        "images": []
    }

    def _parse(parts):
        content_type = parts.get_content_type()

        if content_type.startswith('multipart'):
            for part in parts.get_payload():
                _parse(part)
        elif content_type == 'text/plain':
            content['text_body'] = parts.get_payload(decode=True).decode(parts.get_content_charset())
        elif content_type == 'text/html':
            html_coding = parts.get_content_charset()
            content['html_body'] = parts.get_payload(decode=True).decode(html_coding)
            content['html_encoding'] = html_coding
        elif content_type.startswith('image'):
            filename = decode_mail_header(parts.get_filename())[0]
            content["images"].append({
                "filename": filename,
                "content_id": parts.get("Content-id")[1:-1],
                "content_type": parts.get_content_type(),
                "content": parts.get_payload(decode=True)
            })
        else:
            filename = decode_mail_header(parts.get_filename())[0]
            # 其它情况下，只可能为附件或者嵌入html页面内的元素，或者通过content_type是否为"application"开头来判断是否为附件
            if "attachment" in parts.get("Content-Disposition", []):
                content["attachments"].append({
                    "filename": filename,
                    "content": parts.get_payload(decode=True)
                })
            else:
                raise NotImplemented(f"Don't know how to deal with type: {content_type}")

    _parse(raw_mail)

    return content


def image_to_base64(image: Union[str, bytes], encoding):
    """
    将图片转换成base64编码
    :param image: 字节码或者字符串，如果是字符串，表示图片的路径
    :param encoding: 字符串编码
    :return: base64编码字符串
    """

    if isinstance(image, str):
        image = Path(image).read_bytes()
    image_base64 = base64.b64encode(image)
    image_base64 = image_base64.decode(encoding)
    return image_base64


def imap_utf7_encode(text: str):
    """将字符串转换成imap支持的utf7编码，并将+号替换成&号"""
    return text.encode('utf7').replace(b'+', b'&')


def imap_utf7_decode(bytes_: bytes):
    """将imap的字节编码转换成字符串，imap是utf7格式，并将+号替换成&号"""
    return bytes_.replace(b'&', b'+').decode('utf7')




import email
import re
from email import generator
from pathlib import Path
from datetime import datetime
from typing import TYPE_CHECKING
from .utils import decode_mail_header, parse_raw_mail, image_to_base64

VALID_FLAGS = ['Seen', 'Flagged', 'Answered', 'Draft', 'Deleted', 'Recent']

if TYPE_CHECKING:
    from .folder import Folder


class Mail:
    def __init__(self, mail_id: int | str, folder: 'Folder'):
        self.folder = folder
        self.box = folder.box
        self.server = folder.server
        self.mail_id = str(mail_id)
        self._raw_mail = None
        self._content = None
        self._headers = None

    def __getattr__(self, item):
        return getattr(self.raw_mail, item)

    def _fetch(self, command):
        """根据指令获取邮件内容"""
        # imap fetch的第二个参数是用括号括起来的1个或者多个指令，比如(RFC822), (RFC822 FLAGS)
        # 返回的data是一个列表，如果列表元素是一个元组，则元组的第二个元素是邮件实体
        # 当指定的mail_id不正确的时候，返回的值为('OK', [None])
        typ, data = self.server.fetch(self.mail_id, command)

        outputs = []

        for resp in data:
            if isinstance(resp, tuple):
                outputs.append(resp[1].decode("utf-8"))
            if isinstance(resp, bytes):
                outputs.append(resp.decode("utf-8"))

        return outputs

    @property
    def content(self):
        """返回一个字典"""
        if self._content is None:
            self._content = parse_raw_mail(self.raw_mail)
        return self._content

    @property
    def raw_mail(self):
        """返回邮件原始内容，是email包的Message对象"""
        if self._raw_mail is None:
            data = self._fetch("(RFC822)")
            self._raw_mail = email.message_from_string(data[0])
        return self._raw_mail

    @property
    def headers(self):
        """获取邮件元信息"""
        if self._headers is None:
            self._headers = {k.lower(): v for k, v in self.raw_mail.items()}
        return self._headers

    def _get_mail_info(self, key):
        """获取邮件头中指定的值"""
        value = self.headers.get(key)
        if value:
            value = ''.join(decode_mail_header(value))
        return value

    @property
    def subject(self):
        return self._get_mail_info("subject")

    @property
    def sender(self):
        return self._get_mail_info("sender")

    @property
    def from_(self):
        return self._get_mail_info("from")

    @property
    def to(self):
        return self._get_mail_info("to")

    @property
    def date(self):
        d, b, y, t, z = self.headers["date"].split(',')[1].split()[:5]
        dt = datetime.strptime(f"{d} {b} {y} {t} {z}", "%d %b %Y %H:%M:%S %z")
        return dt.astimezone().strftime("%Y-%m-%d %H:%M:%S")

    @property
    def text_body(self):
        """获取邮件的文本内容"""
        return self.content.get("text_body")

    @property
    def html_body(self):
        """获取邮件html的内容"""
        return self.content.get("html_body")

    def save_html(self, save_path='.'):
        """保存邮件为html文件，图片编码成base64格式嵌入"""
        html_body = self.html_body

        if not html_body:
            raise ValueError("mail has no html body")

        save_path = Path(save_path)
        Path.mkdir(save_path, exist_ok=True)
        html_path = save_path / f'{self.mail_id}.html'

        for image in self.content["images"]:
            content_id = image['content_id']
            content_type = image['content_type']
            content = image_to_base64(image['content'], self.content['html_coding'])
            base64_image = f"data:{content_type};base64,{content}"
            html_body = html_body.replace(f"cid:{content_id}", base64_image)

        html_path.write_text(html_body, self.content["html_encoding"])

    @property
    def attachments(self):
        """返回一个列表，元素是字典，字典的键是附件名称，值是附件二进制内容"""
        return self.content["attachments"]

    def save_attachments(self, save_path='.'):
        """保存所有附件，返回附件路径组成的列表"""
        save_path = Path(save_path)
        Path.mkdir(save_path, exist_ok=True)

        pathes = []

        for attachment in self.attachments:
            filepath = save_path / Path(attachment["filename"])
            filepath.write_bytes(attachment["content"])
            pathes.append(str(filepath))

        return pathes

    def save(self, filename=None):
        """将邮件保存为eml文件"""
        if filename is None:
            filename = f"{self.mail_id}.eml"

        filename = Path(filename)

        if filename.suffix != '.eml':
            raise ValueError("file suffix must be .eml")

        with open(filename, 'wt') as eml:
            gen = generator.Generator(eml)
            gen.flatten(self.raw_mail)

    @property
    def flags(self) -> list[str]:
        # self._fetch('FLAGS')的结果为['1 (FLAGS (\\Seen \\Flagged))']或者['1 (FLAGS ())']
        # 转换成'1 FLAGS \\Seen \\Flagged'或者'1 FLAG  '
        flags = self._fetch('FLAGS')[0].replace('(', '').replace(')', '').replace('\\', '')
        # split不加参数会去掉结果中包含的''
        flags = flags.split()[2:]
        return flags

    # 把flag设置为mail的特性容易和text_body等属性造成混淆，所以统一通过add_flags,set_flags,remove_flags来设置标志
    def _store_flags(self, command: str, flags: str):
        """设置邮件标志通用方法"""
        try:
            flags = re.split(r',|\s+', flags)
        except TypeError:
            pass

        flags = [flag.capitalize() for flag in flags]

        for flag in flags:
            if flag not in VALID_FLAGS:
                raise ValueError(f'{flag} is not a valid flag.')

        flags = ' '.join([rf'\{flag}' for flag in flags])
        self.server.store(self.mail_id, command, flags)

    def set_flags(self, flags):
        """设置邮件标识，可用标识有seen, flagged, answered, draft, deleted"""
        self._store_flags('FLAGS', flags)

    def add_flags(self, flags):
        """添加邮件标识"""
        self._store_flags('+FLAGS', flags)

    def remove_flags(self, flags):
        """移除邮件标识"""
        self._store_flags('-FLAGS', flags)

    def move_to(self, folder_name):
        """将邮件移动到指定文件夹"""
        self.server.copy(self.mail_id, self.box._folders[folder_name])
        self.add_flags('deleted')

    def __repr__(self):
        return f"Mail<{self.mail_id}>"



from collections import UserList
from typing import TYPE_CHECKING
from .email import Mail

if TYPE_CHECKING:
    from .server import ImapEasyBox


class FoldList(UserList):
    def __getitem__(self, item):
        val = super().__getitem__(item)
        if isinstance(val, Folder):
            val.server.select(val.folder_name)
        return val


class Folder:
    def __init__(self, folder_name: str, box: 'ImapEasyBox'):
        self.box = box
        self.server = box.server
        self.folder_name = folder_name

    @property
    def mails(self):
        return self.search('ALL')

    def search(self, query=None, *, encoding=None, **kwargs):
        """根据条件搜索邮件"""
        if query is None:
            query = self._format_search_query(kwargs)

        if encoding:
            query = query.encode(encoding)

        typ, data = self.server.search(encoding, query)

        if typ != 'OK':
            raise RuntimeError(data[0].decode("ascii"))

        mail_ids = data[0].decode('ascii').split()
        return [Mail(i, self) for i in mail_ids]

    @staticmethod
    def _format_search_query(kwargs):
        """根据传入search方法的关键字参数构造原生的搜索条件字符串"""
        criteria = []

        for field, value in kwargs.items():
            field = field.rstrip('_').upper()

            if isinstance(value, str):
                criteria.append(f"({field} {value})")

            if isinstance(value, bool):
                if value:
                    criteria.append(f"({field})")
                else:
                    criteria.append(f"(NOT {field})")

        query = f"({' '.join(criteria)})"
        return query

    def rename(self, folder_name):
        """更改文件夹名称"""
        self.box.rename_folder(self.folder_name, folder_name)
        self.folder_name = folder_name

    def delete(self):
        """删除文件夹"""
        self.box.delete_folder(self.folder_name)
        self.folder_name = None
        self.box = None
        self.server = None

    def __repr__(self):
        return f"Folder<{self.folder_name}>"
